;; Nous utilisons comme representation des noeuds la structure suivante:
;; ( ( label1 (table1 (liste de peres1)))
;;   ( label2 (table2 (liste de peres2)))
;;   ...
;;   ( labelN (tableN (liste de peresN))))


;; Nous allons inclure tout cela dans une frame : **RB-Liaison**
;; On adjoint  cette frame une liste contenant le nom de tout les noeuds
;; **RB-Var** pour permettre une rcursion sur les noeuds.

(declare (usual-integrations)) ;; Utile pour une meilleur compilation

(define tosave '()) ;; Utilis pour sauver un rseau.

(define make-binding cons)
  
(define binding-var car)

(define binding-val cdr)

(define first-binding car)

(define other-binding cdr)

(define adjoin-binding cons)

(define binding-in-frame
  (lambda (var frame)
    (cond ((null? frame) #f)
	  ((equal? var (binding-var (first-binding frame))) (first-binding frame))
	  (else (binding-in-frame var (other-binding frame))))))

(define extend-frame
  (lambda (var value)
    (set! **RB-Liaison** (adjoin-binding (make-binding var value)
					 **RB-Liaison**))))

(define init 
  (lambda ()
    (begin
      (set! **RB-Liaison** '())
      (set! **RB-Var** '())
      (display "Initialization done"))))

(define **RB-Liaison** '())
(define **RB-Var** '())

(define make-node
  (lambda (label table list-of-father states associated-node)
    (extend-frame label
		  (list table list-of-father states associated-node))
    (set! **RB-Var** (cons label **RB-Var**))
    (display "Nodes in bayesian net. : ")
    (display **RB-Var**)
    (set! tosave (cons **RB-VAR** **RB-Liaison**))))

(define table?
  (lambda (label)
    (let ((b (binding-in-frame label **RB-Liaison**)))
      (cadr b))))

(define father?
  (lambda (label)
    (let ((b (binding-in-frame label **RB-Liaison**)))
      (caddr b))))

(define states?
  (lambda (label)
    (let ((b (binding-in-frame label **RB-Liaison**)))
      (cadddr b))))

(define associated-node?
  (lambda (label)
    (let ((b (binding-in-frame label **RB-Liaison**)))
      (car (cddddr b)))))

(define set-associated-node!
  (lambda (label new-value)
    (let ((b (cddddr (binding-in-frame label **RB-Liaison**))))
      (set-car! b new-value))))

;; A present on peut passer  la mise en oeuvre des fonctions 
;; agissant sur les graphes.

;; Rappelons tout d'abord les differentes tapes  suivre pour transformer
;; un rseau baysien en un arbre de jonction.
;; 1) Moralisation du graphe
;; 2) Triangulation du graphe moral
;; 3) Cration du graphe de regroupement
;; 4) Construction de l'arbre de jonction. ( Algorithme de Kruskal )


;; Commenons par la MORALISATION
;; ------------------------------

;; Representation du graphe moral :
;; L'on reprsente le graphe par l'ensemble des liens.

;; Stratgie :
;; 1) Parcourir la frame **RB-Liaison** pour obtenir la liste de tout les pres
;; 2) Lier entre eux les parents d'un noeud

(define list-of-fathers?
  (lambda (lst)
    (cond ((null? lst) '())
	  (else (cons (cons (car lst) (father? (car lst)))
		      (list-of-fathers? (cdr lst)))))))

(define list-of-fathers 
  (lambda ()
    (list-of-fathers? **RB-Var**)))

(define moralization
  (lambda (lst)
    (cond ((null? lst) '())
	  (else (let ((var (caar lst))
		      (fat (cdar lst))
		      (mor (moralization (cdr lst))))
		  (append (new-bind var fat mor)
			  mor))))))

(define new-bind
  (lambda (var fat lst)
    (let ((newlist (mk-list-of-link var fat)))
      (new-bindbis newlist lst))))

(define mk-list-of-link
  (lambda (origin fathers)
    (let ((old (link origin fathers))
	  (new (link-fath fathers)))
      (append old new))))

(define link
  (lambda (origin fath)
    (cond ((null? fath) '())
	  (else (cons (list origin
			    (car fath))
		      (link origin (cdr fath)))))))

(define link-fath
  (lambda (lst)
    (cond ((null? lst) '())
	  (else (append (link (car lst) (cdr lst))
			(link-fath (cdr lst)))))))

(define new-bindbis
  (lambda (lst1 lst2)
    (cond ((null? lst2) lst1)
	  ((null? lst1) '())
	  ((member-spec? (car lst1) lst2) (new-bindbis (cdr lst1) lst2))
	  (else (cons (car lst1)
		      (new-bindbis (cdr lst1) lst2))))))

(define member-spec?
  (lambda (link lst)
    (or (member? link lst)
	(member? (list (cadr link) (car link)) lst))))

(define member?
  (lambda (link lst)
    (cond ((null? lst) '())
	  ((equal? link (car lst)) #t)
	  (else (member? link (cdr lst))))))


;; Passons maintenant  la TRIANGULARISATION du graphe
;; ---------------------------------------------------

;; Nous appliquons l'heuristique suivante :

;; Tant que le graphe n'est pas vide 
;;         - S'il existe un noeud simple, on le supprime
;;         - S'il n'existe pas de noeud simple, on prend le noeud qui a le moins de voisin
;;           on connecte entre eux les voisins non lis puis on supprime le noeud.

;; Noeud simple : noeud dont tout les voisins sont connects entre eux

;; Nous allons commencer par faire une fonction qui retourne la liste des voisins
;; d'un noeud

(define voisinage
  (lambda (noeud lst)
    (cond ((null? lst) '())
	  ((member? noeud (car lst)) (cons (if (eq? (caar lst) noeud)
					       (cadar lst)
					       (caar lst))
					   (voisinage noeud (cdr lst))))
	  (else (voisinage noeud (cdr lst))))))

;; Maintenant nous allons faire un predicat : simple? qui est vrai
;; si le noeud est simple.

(define simple?
  (lambda (node lst)
    (connected? (voisinage node lst) lst)))

(define connected?
  (lambda (lov lst)
    (cond ((null? lov) #t)
	  (else (and (connected-1st? (car lov) (cdr lov) lst)
		     (connected? (cdr lov) lst))))))

(define connected-1st?
  (lambda (node voisin lst)
    (cond ((null? voisin) #t)
	  ((not (member-spec? (list node (car voisin)) lst)) '())
	  (else (connected-1st? node (cdr voisin) lst)))))

;; Simple-node est une fonction retournant la liste des noeuds simple du graphe

(define Simple-node
  (lambda (grph)
    (Simple-nodebis (list-of-node grph) grph)))

(define Simple-nodebis
  (lambda (lon grph)
    (cond ((null? lon) '())
	  ((simple? (car lon) grph) (cons (car lon) (Simple-nodebis (cdr lon) grph)))
	  (else (Simple-nodebis (cdr lon) grph)))))

(define list-of-node
  (lambda (grph)
    (list-of-node-iter grph '())))

(define list-of-node-iter
  (lambda (grph iter)
    (cond ((null? grph) iter)

	  ((not (or (member? (caar grph) iter)
		    (member? (cadar grph) iter))) 
	   (list-of-node-iter (cdr grph) (append (car grph) iter)))

	  ((not (member? (caar grph) iter)) 
	   (list-of-node-iter (cdr grph) (append (list (caar grph)) iter)))

	  ((not (member? (cadar grph) iter)) 
	   (list-of-node-iter (cdr grph) (append (list (cadar grph)) iter)))

	  (else (list-of-node-iter (cdr grph) iter)))))

;; Least-neigh est une fonction retournant le noeud ayant le moins de voisins.

(define length
  (lambda (lst)
    (cond ((null? lst) 0)
	  (else (+ 1 (length (cdr lst)))))))

(define Least-neigh
  (lambda (grph)
    (let ((lon (map (lambda (x) (cons x (length (voisinage x grph)))) (list-of-node grph))))
      (car (trie (cdr lon) (car lon))))))

(define trie
  (lambda (lst iter)
    (cond ((null? lst) iter)
	  ((> (cdr iter)
	      (cdar lst)) (trie (cdr lst) (car lst)))
	  (else (trie (cdr lst) iter)))))

;; Suppress est une fonction qui supprime les liens ayant un noeud comme extremite
;; elle retourne 2 listes : - L'une est la liste des noeuds supprims.
;;                          - L'autre est la liste restante. 

(define Suppress
  (lambda (node grph)
    (let ((sup (if (null? grph)
		   (cons '() '())
		   (suppress node (cdr grph)))))
      (cond ((null? grph) sup)
	    ((member? node (car grph)) (cons (cons (car grph) (car sup))
					     (cdr sup)))
	    (else (cons (car sup)
			(cons (car grph) (cdr sup))))))))

;; Passons maintenant  une fonction qui cre les liens manquant  un noeud non-simple

	
(define add-link
  (lambda (node grph)
    (add-linkbis (all-link (voisinage node grph)) grph)))

(define add-linkbis
  (lambda (lnk grph)
    (cond ((null? lnk) grph)
	  ((member-spec? (car lnk) grph) (add-linkbis (cdr lnk) grph))
	  (else (add-linkbis (cdr lnk) (cons (car lnk) grph))))))

(define all-link
  (lambda (lst)
    (cond ((null? lst) '())
	  ((null? (cdr lst)) (display "Ne devrait jamais arriver"))
	  ((null? (cddr lst)) (list lst))
	  (else (append (map (lambda (x) (list (car lst) x)) (cdr lst))
			(all-link (cdr lst)))))))

;; Nous pouvons maintenant passer  la triangularisation du graphe.

(define triangularization
  (lambda (grph)
    (triang-iter grph (cons '() '()))))

(define triang-iter
  (lambda (grph iter)
    (cond ((null? grph) iter)
	  (else (let ((smpl (Simple-node grph)))
		  (cond (smpl (let ((sup (Suppress (car smpl) grph)))
				(triang-iter (cdr sup) (cons (cons (cons (car smpl) (voisinage (car smpl) grph))
								   (car iter))
							     (append (car sup) (cdr iter))))))
			(else (triang-iter (add-link (Least-neigh grph) grph) iter))))))))


;; Cration du graphe de REGROUPEMENT
;; ----------------------------------

;; Un graphe de regroupement W est un graphe dont chaque lment u est
;; une partie d'un ensemble V et tel que l'union de tout les u soit gale
;;  V.

;; Le regroupement choisi sera celui des voisins d'un noeud lors sa suppression
;; dans l'algorithme de triangularisation.
;; C'est pourquoi triangularization retourne 2 valeurs : le graphe (purement  titre indicatif)
;; et les clusters (la seule chose qui nous interesse en fait).

;; En fait la structure retourne par clusterize n'est pas un graphe  proprement parler mais
;; uniquement les noeuds composant le graphe.

(define inter
  (lambda (lst1 lst2)
    (cond ((null? lst1) '())
	  ((member? (car lst1) lst2) (cons (car lst1) (inter (cdr lst1) lst2)))
	  (else (inter (cdr lst1) lst2)))))

(define union
  (lambda (lst1 lst2)
    (cond ((null? lst1) lst2)
	  ((member? (car lst1) lst2) (union (cdr lst1) lst2))
	  (else (append (car lst1)
			(union (cdr lst1) lst2))))))

(define included?
  (lambda (lst1 lst2)
    (cond ((null? lst1) #t)
	  (else (and (member? (car lst1) lst2)
		     (included? (cdr lst1) lst2))))))

(define ensequal?
  (lambda (lst1 lst2)
    (and (included? lst1 lst2) (included? lst2 lst1))))

;; Il ne sert  rien d'avoir un noeud qui a la mme table que le sparateur => on le
;; supprime. Ex : (clusterize ((i h) (g h i) (g f i) (f h))) ==> ((g h i) (g f i) (f h))
;; ce qui nous permet de supprimer toute information redondante inutile.
;; NB : On ne peut pas supprimer toute redondance puisque c'est l'essence mme du RB.

(define clusterize
  (lambda (lst)
    (clusterize-iter lst '())))

(define clusterize-iter
  (lambda (lst acc)
    (cond ((null? lst) acc)
	  ((or2 (map (lambda (x) (included? (car lst) x)) (cdr lst))) (clusterize-iter (cdr lst) acc))
	  (else (clusterize-iter (cdr lst) (cons (car lst) acc))))))

(define or2
  (lambda (lst)
    (cond ((null? lst) #f)
	  ((car lst) #t)
	  (else (or2 (cdr lst))))))

;; Creation de l'arbre de jonction
;; -------------------------------

;; Il ne reste plus qu'a crer l'arbre de jonction ... Et ceci en appliquant l'algorithme de Kruskal

;;        Algorithme de Kruskal :
;;           * Ordonner tous les arcs possibles par poids dcroissant.
;;           * Initialiser E  ()
;;           * Pour chaque arc e
;;                  - Si le graphe obtenu en ajoutant e  E n'a pas de cycle alors l'ajouter.

;; On sait que tout graphe triangul possde un arbre de jonction, de plus on sait qu'un arbre de jonction
;; est facilement trouv grace  l'algorithme de Kruskal en prenant comme fonction de poids : |(inter u v)|

;; Creation de la liste des arcs

(define list-of-arcs
  (lambda (lst)
    (cond ((null? lst) '())
	  (else (append (add-arc (car lst) (cdr lst))
			(list-of-arcs (cdr lst)))))))

(define add-arc
  (lambda (orig lst)
    (cond ((null? lst) '())
	  (else (cons (list orig (car lst))
		      (add-arc orig (cdr lst)))))))

;; Calcul du poids

(define poids
  (lambda (grph)
    (map (lambda (x) (cons (length (inter (car x) (cadr x)))
			   x))
	 grph)))

;; Tri de la liste

(define tri-dec
  (lambda (lst)
    (cond ((null? lst) '())
	  (else (tri-dec-iter (cdr lst) (list (car lst)))))))

(define tri-dec-iter
  (lambda (lst acc)
    (cond ((null? lst) acc)
	  ((>= (caar lst) (caar acc)) (tri-dec-iter (cdr lst)
						    (cons (car lst)
							  acc)))
	  (else (tri-dec-iter (cdr lst) (add-to-place (car lst) acc))))))

(define add-to-place
  (lambda (tta lst)
    (cond ((null? lst) (list tta))
	  ((>= (car tta) (caar lst)) (cons tta lst))
	  (else (cons (car lst)
		      (add-to-place tta (cdr lst)))))))

;; Le prediact add-cycle?

;; On a un graphe sans cycle et l'on doit vrifier si le fait d'ajouter un lien cre un cycle.
;; On doit donc verifier si les deux noeuds nouvellement connects ne le sont pas par ailleurs 
;; dans le graphe.

(define add-cycle?
  (lambda (lnk grph)
    (cond ((null? grph) '())
	  (else (member? (car lnk) (all-connection (cadr lnk) grph))))))

(define all-connection
  (lambda (node grph)
    (cond ((null? grph) '())
	  ((null? node) '())
	  (else (append (voisinage node grph)
			(apply append
			       (map (lambda (x)
				      (all-connection x (cdr (suppress node grph))))
				    (voisinage node grph))))))))

;; Et enfin l'algorithme de Kruskal

(define Kruskal
  (lambda (lst)
    (Kruskal-iter lst '())))

(define Kruskal-iter
  (lambda (lst acc)
    (cond ((null? lst) acc)
	  ((add-cycle? (car lst) acc) (kruskal-iter (cdr lst) acc))
	  (else (kruskal-iter (cdr lst) (cons (car lst) acc))))))


;; La fonction qui opre les oprations de cration du Junction tree

(define Junction
  (lambda ()
    (Kruskal (map cdr (tri-dec (poids (list-of-arcs (clusterize (car (triangularization (moralization (list-of-fathers))))))))))))

